#ifndef _ITRLEDA_DD_H_
#define _ITRLEDA_DD_H_

/* -----------------------------------------------------------------------
 *
 *  Table data definitions for ITR
 *
 *  This code is heavily based on VDSLEDA by R. Wanke et. al.
 *
 *
 *  The central data structure is a read-out chip.
 *  The following relations hold:
 *
 *  FED(DAQ) 1 <-> 1 ITR_Fed 1 <-> m ITR_Chip m <-> 1 ITR_Chamber  1 <-> 1 ITR_NEDE
 *                                            1 <-> 1 ITR_Chanstat
 *                                            1 <-> 1 ITR_Ped 
 *
 *  There are two possible pedestal (and noise) tables:
 *    - ITR_Ped:      One cell independent pedestal for 128 cells 
 *    - ITR_PedHelix: 141 Helix pipeline cells for 128 cells 
 *  => polymorphism is used in the cell <-> pedestal table relation.
 *
 * ITR_Fed should disappear soon and be replaced by DAQ_Fed !
 * ----------------------------------------------------------------------- */
/* Important:
 * - Underscores in table names are translated to '/' in Mizzi Tables.
 * - The hash table dimension should be of the order of magnitude of table
 *   entries and a prime number to increase performances.
 * - The key tuple have been fixed to be of five fields
 *   at maximum ( SetLedaKey(5) (...) ), but can be easily increased if needed.
 * - SetLedaRelation fixes the corrispondence among the relation containers
 *   of the related tables. The example above reads: table
 *   'ECAL_Module' is relatable through its 'containedCells' relation 
 *   container to table 'ECAL_Cell' through its 'containerModule' 
 *   relation container. More than one corrispondence can be established
 *   with the same relation container (see for example ECAL_Board dd below).
 *   The maximum number of corrispondences have been fixed
 *   to 32 ( SetLedaRelation(32) (...) ), but can be increased.
 */

/***********************************************************************
 * CHANGES
 *
 * Apr 04, 2001 (wg):  introduce new table ITR/EffRes in ITR_MONITOR containing 
 *                     efficiencies and resolutions per chamber.  
 *                     NO relations to ITR/Chamber to keep backward compatibility
 ***********************************************************************/


/* LEDA include */
#ifdef __cplusplus
#include "ledaC++.hh"
#else
#include "ledaC.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif


/* define some constants - will always be the same   */
/* (But unfortunately we have to compute them below) */
#define NCHIPCHAN  128
#define NCELLHELIX 141

/* lengths of names */
#define LCHAMBERNAME     50

/* ============================================================= */

/* ITR Chips
 * ========= */
DefineLedaTable (ITR_Chip, 1153,
		 const DBushort   id;         /* Key */
		 DBubyte          ncellped;   /* # of cells for peds - may be 1 */
		 DBubyte          ncellnoise; /* # of cells for noise */
		 DBbyte           status;     /* chip status */
		 DBubyte          pipelen;    /* length of pipeline */
		 const DBrelation fed;
		 const DBrelation chamber;
		 const DBrelation chanstat;
		 const DBrelation noise;
		 const DBrelation pedestals;
		 );

SetLedaKey(1) (ITR_Chip,
	       DBushort, id
	       );

SetLedaRelation(7) (ITR_Chip, 
		     fed,       ITR_Fed,        chip,
		     chamber,   ITR_Chamber,    chip,
		     chanstat,  ITR_Chanstat,   chip,
		     noise,     ITR_Noise,      chip, /* hurray - here we use */
		     noise,     ITR_NoiseHelix, chip, /* polymorphism         */
		     pedestals, ITR_Ped,        chip, /* hurray - here we use */
		     pedestals, ITR_PedHelix,   chip  /* polymorphism         */
		     );

DefineLedaRelationTable (ITR_Chip,
			 ITR_Fed,        ITR_ChipFed,
			 ITR_Chamber,    ITR_ChipChamber,
			 ITR_Chanstat,   ITR_ChipChanstat,
			 ITR_Noise,      ITR_ChipNoise,
			 ITR_NoiseHelix, ITR_ChipNoiseHlx,
			 ITR_Ped,        ITR_ChipPed,
			 ITR_PedHelix,   ITR_ChipPedHlx,
			 );

/* ============================================================= */

/* ITR FEDs
 * ========= */
DefineLedaTable (ITR_Fed, 577,
		 const DBushort   id;       /* Key */
		 DBubyte          vmeid;
		 DBubyte          board;
		 DBubyte          sharc;
		 DBubyte          link;
		 const DBrelation chip;
		 const DBrelation daq_fed;
		);

SetLedaKey(1) (ITR_Fed, 
	       DBushort, id
	       );

SetLedaRelation(2) (ITR_Fed, 
		    chip,    ITR_Chip,  fed,
		    daq_fed, DAQ_Fed,   vds_fed
		    );

DefineLedaRelationTable (ITR_Fed,
			 ITR_Chip,  ITR_ChipFed,
			 DAQ_Fed,   ITR_DaqFed
			 );


/* ============================================================= */

/* ITR Chambers
 * ========= */
DefineLedaTable (ITR_Chamber, 191,
		 const DBushort   id;         /* Key */
		 DBascii          name[LCHAMBERNAME];
		 DBbyte           align_flag; /* DataValid Offset for this chamber   */
		 DBbyte           status;     /* status (eg for holmes)              */
                 DBubyte          type;       /* type of chamber 1..3                */
		 const DBrelation chip;       /* 1:m relation for readout chips      */
		 const DBrelation align;      /* alignment table                     */
		 const DBrelation tonede;     /* geometry information from NEDE bank */
		 );

SetLedaConstant(ITR_Chamber,
		    LCHAMBERNAME
		);

SetLedaKey(1) (ITR_Chamber,
	       DBushort,   id
	       );

SetLedaRelation(3) (ITR_Chamber, 
		    chip,     ITR_Chip,  chamber,
		    align,    ITR_Align, chamber,
		    tonede,   ITR_nede,  tochamber
		    );

DefineLedaRelationTable (ITR_Chamber,
			 ITR_Chip,      ITR_ChipChamber,
			 ITR_Align,     ITR_ChAlign,
			 ITR_nede,      ITR_cham_R_nede
			 );

/* ============================================================= */

/* ITR Cell Independent Channel Status
 * =================================== */
DefineLedaTable (ITR_Chanstat, 1153,
		 const DBushort   chipid;   /* Key */
		 DBubyte          status[NCHIPCHAN];
		 const DBrelation chip;
		 );

SetLedaConstant(ITR_Chanstat,
		    NCHIPCHAN 
		);

SetLedaKey(1) (ITR_Chanstat,
	       DBushort, chipid
	       );

SetLedaRelation(1) (ITR_Chanstat, 
		    chip, ITR_Chip, chanstat
		    );

DefineLedaRelationTable (ITR_Chanstat,
			 ITR_Chip, ITR_ChipChanstat
			 );

/* ============================================================= */

/* ITR Cell Independent Noise
 * ========================== */
DefineLedaTable (ITR_Noise, 1153,
		 const DBushort   chipid;   /* Key */
		 DBfloat          rawvar[NCHIPCHAN];
		 DBfloat          subvar[NCHIPCHAN];
		 const DBrelation chip;
		 );

SetLedaConstant(ITR_Noise,
		    NCHIPCHAN
		);

SetLedaKey(1) (ITR_Noise,
	       DBushort, chipid
	       );

SetLedaRelation(1) (ITR_Noise, 
		    chip, ITR_Chip, chanstat
		    );

DefineLedaRelationTable (ITR_Noise,
			 ITR_Chip, ITR_ChipNoise
			 );

/* ITR Cell Dependent Noise for Helixes
 * ==================================== */
DefineLedaTable (ITR_NoiseHelix, 1153,
		 const DBushort   chipid;   /* Key */
		 DBfloat          rawvar[NCHIPCHAN][NCELLHELIX];
		 DBfloat          subvar[NCHIPCHAN][NCELLHELIX];
		 const DBrelation chip;
		 );

SetLedaConstant(ITR_NoiseHelix,
		    NCHIPCHAN, NCELLHELIX
		);

SetLedaKey(1) (ITR_NoiseHelix,
	       DBushort, chipid
	       );

SetLedaRelation(1) (ITR_NoiseHelix, 
		    chip, ITR_Chip, chanstat
		    );

DefineLedaRelationTable (ITR_NoiseHelix,
			 ITR_Chip, ITR_ChipNoiseHlx
			 );

/* ============================================================= */

/* ITR Cell Independent Pedestals
 * ============================ */
DefineLedaTable (ITR_Ped, 1153,
		 const DBushort   chipid;   /* Key */
		 DBshort          pedestal[NCHIPCHAN];
		 const DBrelation chip;
		 );

SetLedaConstant(ITR_Ped,
		    NCHIPCHAN
		);

SetLedaKey(1) (ITR_Ped,
	       DBushort, chipid
	       );

SetLedaRelation(1) (ITR_Ped, 
		    chip, ITR_Chip, pedestals
		    );

DefineLedaRelationTable (ITR_Ped,
			 ITR_Chip, ITR_ChipPed
			 );

/* ITR Cell Dependent Pedestals for Helixes
 * ======================================== */
DefineLedaTable (ITR_PedHelix, 1153,
		 const DBushort   chipid;   /* Key */
		 DBshort          pedestal[NCHIPCHAN][NCELLHELIX];
		 const DBrelation chip;
		 );

SetLedaConstant(ITR_PedHelix,
		    NCHIPCHAN, NCELLHELIX
		);

SetLedaKey(1) (ITR_PedHelix,
	       DBushort, chipid
	       );

SetLedaRelation(1) (ITR_PedHelix, 
		    chip, ITR_Chip, pedestals
		    );

DefineLedaRelationTable (ITR_PedHelix,
			 ITR_Chip, ITR_ChipPedHlx
			 );

/* ============================================================= */

/* ITR Alignment table 
 * =================== */

#include "itralign.h"


/* A Copy of the ITr entries in the ARTE table NEDE */
/*==================================================*/
#include "itrdb_nede.h"



/* map of ITR connection to FLT TFU boards */
/*=========================================*/
DefineLedaTable  (ITR_TFU, 191,
		  DBushort   crate;  
		  DBushort   slot;
		  DBushort   recv;
		  DBint      serial;
		  DBint      status;
		  const DBrelation chamber;
		  );
 
SetLedaKey(3)     (ITR_TFU, 
		   DBushort, crate,
		   DBushort, slot,
		   DBushort, recv
		   );

SetLedaRelation (1) (ITR_TFU,
		     chamber, ITR_Chamber, tfu );

DefineLedaRelationTable (ITR_TFU, ITR_Chamber, TFU_R_Chamber);



/* ====================================================================== *
 *   ITR_FEDSLB:  connection between FEDs and SHARCs.                     
 *                This information really should be put into ITR_FED or 
 *                replaced by a relation between ITR_FED and DAQ_FED or so, 
 *                but this would require a completely new table ITR_FED...
 * ====================================================================== */

DefineLedaTable (ITR_FedSlb, 577, 
		 const DBushort   fedid;    /* key */
		 DBushort         slbid;
		 );

SetLedaKey(1)   (ITR_FedSlb, DBushort, fedid);


/* ====================================================================== *
 *   ITR_EffRes:  Efficiency, hit resolution, and some spare numbers
 *                for future use, per chamber.
 * ====================================================================== */

  DefineLedaTable (ITR_EffRes, 191,
		   const DBushort chamber;  /* key */
		   DBfloat        eff;      /* efficiency     */
		   DBfloat        res;      /* hit resolution */
		   DBfloat        spare[14];/* for future use */
		   );

  SetLedaKey(1) (ITR_EffRes, DBushort, chamber);

/* ====================================================================== *
 *   ITR_DVOffset:  Data Valid offset, per chamber.
 * ====================================================================== */

  DefineLedaTable (ITR_DVOffset, 191,
		   const DBushort chamber;  /* key */
		   DBint          offset;   /* DV offset per chamber */
		   );

  SetLedaKey(1) (ITR_DVOffset, DBushort, chamber);


#ifdef __cplusplus
}
#endif

#endif
