#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "grr.h"

#include "aframe.h"
#include "aframelib.h"

#include "itrleda.h"
#include "itrleda_dd.h"
#include "itrleda_structs.h"


static const unsigned short int kNotConnected = 0xFFFF;


/*======================================================================
 * 
 * Routines to fill c-structures from Leda tables and vice versa
 *
 *======================================================================*/

/*----------------------------------------------------------------------
 * fill chamber structure from ITR_Chamber
 *
 * arguments:  itrleda_Chamber_t *chambers    // pointer to pre-allocated array
 *             int            maxChams        // max. space in this array
 *             int            *nCham          // number of entries actually copied
 *
 * returns:    GRR_OK     if enough space was allocated
 *             GRR_FISHY  if more entries than maxChips in ITR_Chip found
 *----------------------------------------------------------------------*/
int ITR_leda_Chamber_structFromTable (itrleda_Chamber_t *chambers,
				      int maxCham, int *nCham)
{
    static char *routine = "ITR_leda_Chamber_structFromTable";
    int count, errflg;
    LedaTableIterator (ITR_Chamber, chamberIt);

    count = 0;
    errflg = GRR_OK;

    for (chamberIt.begin (); chamberIt.valid (); chamberIt.next ()) {
	LedaTableRow (ITR_Chamber) * ledaChamber = chamberIt.get ();
	LedaRelationIterator (ledaChamber->chip, chipIt);
	LedaTableRow (ITR_nede)  * ledaNede  = LedaRelation (ledaChamber->tonede);
	LedaTableRow (ITR_Align) * ledaAlign = LedaRelation (ledaNede->toalign);

	int chipCnt;

	if (count >= maxCham) {
	    errflg = GRR_FISHY;
	    break;
	}

	if (ledaNede == NULL) {
	  aframe_mess (routine, GRR_SERROR, "Nede not found for chamber 0x%04x!?",
		       ledaChamber->id);
	  errflg = GRR_FISHY;
	  continue;
	}

	if (ledaAlign == NULL) {
	  aframe_mess (routine, GRR_SERROR, "Align not found for chamber 0x%04x!?",
		       ledaChamber->id);
	  errflg = GRR_FISHY;
	  continue;
	}

	chambers[count].id = ledaChamber->id;
	strncpy (chambers[count].name, ledaChamber->name,
		 LCHAMBERNAME - 1);

	chambers[count].align_flag = ledaChamber->align_flag;
	chambers[count].status = ledaChamber->status;
	chambers[count].type   = ledaChamber->type;

	/* Initialize the first element of the chip array to something
	 * invalid in case there are no chips connected           	 */
	chambers[count].chip[0] = kNotConnected;

	for (chipIt.begin (), chipCnt = 0; chipIt.valid () && chipCnt < 6;
	     chipIt.next (), chipCnt++) {
	    LedaTableRow (ITR_Chip) * theChip = chipIt.get ();
	    chambers[count].chip[chipCnt] = theChip->id;
	}

	chambers[count].align  = ledaAlign->id;
	chambers[count].tonede = ledaNede ->inede;

	count++;
    }

    *nCham = count;

    return (errflg);
}


/*----------------------------------------------------------------------
 * fill chip structure from ITR_Chip
 *
 * arguments:  itrleda_Chip_t *chips    // pointer to pre-allocated array
 *             int            maxChips  // max. space in this array
 *             int            *nChips   // number of entries actually copied
 *
 * returns:    GRR_OK     if enough space was allocated
 *             GRR_FISHY  if more entries than maxChips in ITR_Chip found
 *----------------------------------------------------------------------*/
int ITR_leda_Chip_structFromTable (itrleda_Chip_t *chips,
				   int maxChips, int *nChips)
{

    int count, errflg;
    LedaTableIterator (ITR_Chip, chipIt);

    count = 0;
    errflg = GRR_OK;

    for (chipIt.begin (); chipIt.valid (); chipIt.next ()) {
	LedaTableRow (ITR_Chip)    * ledaChip    = chipIt.get ();
	LedaTableRow (ITR_Chamber) * ledaChamber = LedaRelation (ledaChip->chamber);
	LedaTableRow (ITR_Fed)     * ledaFed     = LedaRelation (ledaChip->fed);

	if (count >= maxChips) {
	    errflg = GRR_FISHY;
	    break;
	}

	chips[count].id = ledaChip->id;
	chips[count].ncellped = ledaChip->ncellped;
	chips[count].ncellnoise = ledaChip->ncellnoise;
	chips[count].status = ledaChip->status;
	chips[count].pipelen = ledaChip->pipelen;

	chips[count].fed = ledaFed->id;
	chips[count].chamber = ledaChamber->id;

	count++;
    }

    *nChips = count;

    return (errflg);
}


/*----------------------------------------------------------------------
 * fill chanstat structure from ITR_Chanstat
 *
 * arguments:  itrleda_Chanstat_t *chstat    // pointer to pre-allocated array
 *             int            maxChips  // max. space in this array
 *             int            *nChips   // number of entries actually copied
 *
 * returns:    GRR_OK     if enough space was allocated
 *             GRR_FISHY  if more entries than maxChips in ITR_Chanstat found
 *----------------------------------------------------------------------*/

int ITR_leda_Chanstat_structFromTable (itrleda_Chanstat_t *chstat,
				       int maxChips, int *nChips)
{

    int count, errflg, i;
    LedaTableIterator (ITR_Chanstat, chIt);

    count = 0;
    errflg = GRR_OK;

    for (chIt.begin (); chIt.valid (); chIt.next ()) {
	LedaTableRow (ITR_Chanstat) * ledaCh = chIt.get ();

	if (count >= maxChips) {
	    errflg = GRR_FISHY;
	    break;
	}

	chstat[count].chipid = ledaCh->chipid;
	
	for (i = 0; i < NCHIPCHAN; i++) 
	  chstat[count].status[i] = ledaCh->status[i];

	/* memcpy(chstat[count].status, ledaCh->status, NCHIPCHAN); */

	count++;
    }

    *nChips = count;

    return (errflg);
}


/*----------------------------------------------------------------------
 * fill pedestal structure from ITR_Ped
 *
 * arguments:  itrleda_Ped_t *ped       // pointer to pre-allocated array
 *             int            maxChips  // max. space in this array
 *             int           *nChips    // number of entries actually copied
 *
 * returns:    GRR_OK     if enough space was allocated
 *             GRR_FISHY  if more entries than maxChips in ITR_Ped found
 *----------------------------------------------------------------------*/

int ITR_leda_Ped_structFromTable (itrleda_Ped_t *ped,
				  int maxChips, int *nChips)
{

    int count, errflg, i;
    LedaTableIterator (ITR_Ped, pedIt);

    count = 0;
    errflg = GRR_OK;

    for (pedIt.begin (); pedIt.valid (); pedIt.next ()) {
	LedaTableRow (ITR_Ped) * ledaPed = pedIt.get ();

	if (count >= maxChips) {
	    errflg = GRR_FISHY;
	    break;
	}

	ped[count].chipid = ledaPed->chipid;
	
	for (i = 0; i < NCHIPCHAN; i++) 
	  ped[count].pedestal[i] = ledaPed->pedestal[i];

	count++;
    }

    *nChips = count;

    return (errflg);
}


/*----------------------------------------------------------------------
 * fill noise structure from ITR_Noise
 *
 * arguments:  itrleda_Noise_t *noise      // pointer to pre-allocated array
 *             int              maxChips   // max. space in this array
 *             int             *nChips     // number of entries actually copied
 *
 * returns:    GRR_OK     if enough space was allocated
 *             GRR_FISHY  if more entries than maxChips in ITR_Noise found
 *----------------------------------------------------------------------*/

int ITR_leda_Noise_structFromTable (itrleda_Noise_t *noise,
				  int maxChips, int *nChips)
{

    int count, errflg, i;
    LedaTableIterator (ITR_Noise, noiseIt);

    count = 0;
    errflg = GRR_OK;

    for (noiseIt.begin (); noiseIt.valid (); noiseIt.next ()) {
	LedaTableRow (ITR_Noise) * ledaNoise = noiseIt.get ();

	if (count >= maxChips) {
	    errflg = GRR_FISHY;
	    break;
	}

	noise[count].chipid = ledaNoise->chipid;
	
	for (i = 0; i < NCHIPCHAN; i++) {
	  noise[count].rawvar[i] = ledaNoise->rawvar[i];
	  noise[count].subvar[i] = ledaNoise->subvar[i];
	}

	count++;
    }

    *nChips = count;

    return (errflg);
}


/*----------------------------------------------------------------------
 * fill FED structure from ITR_Fed
 *
 * arguments:  itrleda_Fed_t *fed      // pointer to pre-allocated array
 *             int            maxFeds  // max. space in this array
 *             int           *nFeds    // number of entries actually copied
 *
 * returns:    GRR_OK     if enough space was allocated
 *             GRR_FISHY  if more entries than maxChips in ITR_Fed found
 *----------------------------------------------------------------------*/

int ITR_leda_Fed_structFromTable (itrleda_Fed_t *chstat,
				       int maxFeds, int *nFeds)
{

    int count, errflg, chipCnt;
    LedaTableIterator (ITR_Fed, fedIt);

    count = 0;
    errflg = GRR_OK;

    for (fedIt.begin (); fedIt.valid (); fedIt.next ()) {
	LedaTableRow (ITR_Fed) * ledaFed = fedIt.get ();
	LedaRelationIterator (ledaFed->chip, chipIt);

	if (count >= maxFeds) {
	    errflg = GRR_FISHY;
	    break;
	}

	chstat[count].id    = ledaFed->id;
	chstat[count].vmeid = ledaFed->vmeid;
	chstat[count].board = ledaFed->board;
	chstat[count].sharc = ledaFed->sharc;
	chstat[count].link  = ledaFed->link;

	chipCnt = 0;
	for (chipIt.begin(); chipIt.valid() && chipCnt < 2; chipIt.next(), chipCnt ++) {
	  LedaTableRow (ITR_Chip) * theChip = chipIt.get ();
	  chstat[count].chip[chipCnt] = theChip->id;	  
	}

	count++;
    }

    *nFeds = count;

    return (errflg);
}



/*----------------------------------------------------------------------
 * fill align structure from ITR_Align
 *
 * arguments:  itrleda_Align_t *align      // pointer to pre-allocated array
 *             int              maxAlign   // max. space in this array
 *             int             *nAlign     // number of entries actually copied
 *
 * returns:    GRR_OK     if enough space was allocated
 *             GRR_FISHY  if more entries than maxAlign in ITR_Align found
 *----------------------------------------------------------------------*/

int ITR_leda_Align_structFromTable (itrleda_Align_t *align,
				  int maxAlign, int *nAlign)
{

    int count, errflg, i;
    LedaTableIterator (ITR_Align, alignIt);

    count = 0;
    errflg = GRR_OK;

    for (alignIt.begin (); alignIt.valid (); alignIt.next ()) {
	LedaTableRow (ITR_Align) * ledaAlign = alignIt.get ();
	LedaTableRow (ITR_nede)  * ledaNede  = LedaRelation (ledaAlign->tonede);

	if (count >= maxAlign) {
	    errflg = GRR_FISHY;
	    break;
	}

	align[count].id   = ledaAlign->id;
	align[count].nede = (ledaNede) ? ledaNede->inede : -1;
	align[count].redres = ledaAlign->redres;
	for (i = 0; i < 3; i++) {
	  align[count].ang[i]   = ledaAlign->ang[i];
	  align[count].eang[i]  = ledaAlign->eang[i];
	  align[count].delt[i]  = ledaAlign->delt[i];
	  align[count].edelt[i] = ledaAlign->edelt[i];
	  align[count].sloff[i] = ledaAlign->sloff[i];
	  align[count].esloff[i]= ledaAlign->esloff[i];
	}

	count++;
    }

    *nAlign = count;

    return (errflg);
}


/*----------------------------------------------------------------------
 * fill nede structure from ITR_nede
 *
 * arguments:  itrleda_nede_t  *nede      // pointer to pre-allocated array
 *             int              maxNede   // max. space in this array
 *             int             *nNede     // number of entries actually copied
 *
 * returns:    GRR_OK     if enough space was allocated
 *             GRR_FISHY  if more entries than maxAlign in ITR_Align found
 *----------------------------------------------------------------------*/

int ITR_leda_nede_structFromTable (itrleda_nede_t *nede,
				  int maxNede, int *nNede)
{

    int count, errflg, i;
    LedaTableIterator (ITR_nede, nedeIt);

    count = 0;
    errflg = GRR_OK;

    for (nedeIt.begin (); nedeIt.valid (); nedeIt.next ()) {
	LedaTableRow (ITR_nede)  * ledaNede = nedeIt.get ();
	LedaTableRow (ITR_nesl)  * ledaNesl = LedaRelation(ledaNede->tonesl);
	LedaTableRow (ITR_Chamber)*ledaChamber = LedaRelation(ledaNede->tochamber);
	LedaTableRow (ITR_Align )*ledaAlign = LedaRelation(ledaNede->toalign);

	if (count >= maxNede) {
	    errflg = GRR_FISHY;
	    break;
	}

	/* index and relations */
	nede[count].inede   = ledaNede->inede;
	nede[count].tonesl  = ledaNesl ? ledaNesl->inesl : -1;
	nede[count].tochamber = ledaChamber ? ledaChamber->id : kNotConnected;
	nede[count].toalign = ledaAlign ? ledaAlign->id : kNotConnected;

	for (i = 0; i < 2; i++) {
	  nede[count].xfw_db[i]  = ledaNede->xfw_db[i] ;
	  nede[count].xmin_db[i] = ledaNede->xmin_db[i];
	  nede[count].xmax_db[i] = ledaNede->xmax_db[i];
	  nede[count].ymin_db[i] = ledaNede->ymin_db[i];
	  nede[count].ymax_db[i] = ledaNede->ymax_db[i];
	  nede[count].sxmn_db[i] = ledaNede->sxmn_db[i];
	  nede[count].sxmx_db[i] = ledaNede->sxmx_db[i];
	  nede[count].symn_db[i] = ledaNede->symn_db[i];
	  nede[count].symx_db[i] = ledaNede->symx_db[i];
	  nede[count].sec_db[i]  = ledaNede->sec_db[i] ;
	}
	for (i = 0; i < 9; i++)
	  nede[count].rot_db[i] = ledaNede->rot_db[i];

	nede[count].zfw_db  = ledaNede->zfw_db ;
	nede[count].zmin_db = ledaNede->zmin_db;
	nede[count].zmax_db = ledaNede->zmax_db;
	nede[count].acc_db  = ledaNede->acc_db ;
	nede[count].eff_db  = ledaNede->eff_db ;
	nede[count].dod_db  = ledaNede->dod_db ;
	nede[count].pit_db  = ledaNede->pit_db ;
	nede[count].ms_db   = ledaNede->ms_db  ;
	nede[count].nsl_db  = ledaNede->nsl_db ;
	nede[count].lay_db  = ledaNede->lay_db ;
	nede[count].dedx_db = ledaNede->dedx_db;
	nede[count].noise_db= ledaNede->noise_db;
	nede[count].ares_db = ledaNede->ares_db;
	nede[count].superl_db = ledaNede->superl_db ;

	strncpy (nede[count].hwc_db, ledaNede->hwc_db, 12);

	count++;
    }

    *nNede = count;

    return (errflg);
}


/*----------------------------------------------------------------------
 * fill nesl structure from ITR_nesl
 *
 * arguments:  itrleda_nesl_t  *nesl      // pointer to pre-allocated array
 *             int              maxNesl   // max. space in this array
 *             int             *nNesl     // number of entries actually copied
 *
 * returns:    GRR_OK     if enough space was allocated
 *             GRR_FISHY  if more entries than maxAlign in ITR_Align found
 *----------------------------------------------------------------------*/

int ITR_leda_nesl_structFromTable (itrleda_nesl_t *nesl,
				  int maxNesl, int *nNesl)
{

    int count, errflg;
    LedaTableIterator (ITR_nesl, neslIt);

    count = 0;
    errflg = GRR_OK;

    for (neslIt.begin (); neslIt.valid (); neslIt.next ()) {
	LedaTableRow (ITR_nesl)  * ledaNesl = neslIt.get ();
	LedaTableRow (ITR_nede)  * ledaNede = LedaRelation(ledaNesl->tonede);

	if (count >= maxNesl) {
	    errflg = GRR_FISHY;
	    break;
	}

	/* index and relations */
	nesl[count].inesl   = ledaNesl->inesl;
	nesl[count].tonede  = ledaNede ? ledaNede->inede : -1;

	nesl[count].nsl = ledaNesl->nsl;
	nesl[count].mfl = ledaNesl->mfl;
	nesl[count].lay = ledaNesl->lay;
	nesl[count].x1  = ledaNesl->x1;
	nesl[count].x2  = ledaNesl->x2;
	nesl[count].y1  = ledaNesl->y1;
	nesl[count].y2  = ledaNesl->y2;
	nesl[count].z1  = ledaNesl->z1;
	nesl[count].z2  = ledaNesl->z2;
	nesl[count].dx  = ledaNesl->dx;
	nesl[count].dy  = ledaNesl->dy;
	nesl[count].npl = ledaNesl->npl;
	nesl[count].fpl = ledaNesl->fpl;
	nesl[count].ltyp= ledaNesl->ltyp;
	nesl[count].pnede= ledaNesl->pnede;

	strncpy (nesl[count].name, ledaNesl->name, 4);

	count++;
    }

    *nNesl = count;

    return (errflg);
}



/*----------------------------------------------------------------------
 * fill DVOffset structure from ITR_DVOffset
 *
 * arguments:  itrleda_DVOffset_t *chstat    // pointer to pre-allocated array
 *             int             maxOffsets    // max. space in this array
 *             int              *nOffsets    // number of entries actually copied
 *
 * returns:    GRR_OK     if enough space was allocated
 *             GRR_FISHY  if more entries than maxChips in ITR_Chanstat found
 *----------------------------------------------------------------------*/

int ITR_leda_DVOffset_structFromTable (itrleda_DVOffset_t *chstat,
				       int maxOffsets, int *nOffsets)
{

    int count, errflg, i;
    LedaTableIterator (ITR_DVOffset, chIt);

    count = 0;
    errflg = GRR_OK;

    for (chIt.begin (); chIt.valid (); chIt.next ()) {
	LedaTableRow (ITR_DVOffset) * ledaCh = chIt.get ();

	if (count >= maxOffsets) {
	    errflg = GRR_FISHY;
	    break;
	}

	chstat[count].chamber = ledaCh->chamber;
	chstat[count].offset  = ledaCh->offset;

	count++;
    }

    *nOffsets = count;

    return (errflg);
}







/* ======================================================================
 *
 * Functions to fill in Leda Tables from C structures
 *
 * ======================================================================*/

int ITR_leda_Chamber_tableFromStruct (itrleda_Chamber_t *chambers, int nChambers) {
    /* static char *routine = "ITR_leda_Chamber_tableFromStruct"; */
    int count, errflg;
    LedaTableRow(ITR_Chamber) *chamber;

    count = 0;
    errflg = GRR_OK;

    ITR_Chamber->remove->all();

    for (count = 0; count < nChambers; count ++) {
      chamber = ITR_Chamber->add->key ((unsigned short) chambers[count].id);
      
      strncpy (chamber->name, chambers[count].name, LCHAMBERNAME-1);

      chamber->align_flag = chambers[count].align_flag;
      chamber->status     = chambers[count].status;
      chamber->type       = chambers[count].type;
    }

    return (errflg);
}


int ITR_leda_Chip_tableFromStruct (itrleda_Chip_t *chips, int nChips) {
    /* static char *routine = "ITR_leda_Chip_tableFromStruct"; */
    int count, errflg;
    LedaTableRow(ITR_Chip) *chip;

    count = 0;
    errflg = GRR_OK;

    ITR_Chip->remove->all();

    for (count = 0; count < nChips; count ++) {
      chip = ITR_Chip->add->key ((unsigned short) chips[count].id);
      
      chip->status     = chips[count].status;
      chip->ncellped   = chips[count].ncellped;
      chip->ncellnoise = chips[count].ncellnoise;
      chip->pipelen    = chips[count].pipelen;
    }

    return (errflg);
}


int ITR_leda_Chanstat_tableFromStruct (itrleda_Chanstat_t *cstat, int nChips) {
    /* static char *routine = "ITR_leda_Chanstat_tableFromStruct"; */
    int count, errflg, i;
    LedaTableRow(ITR_Chanstat) *chanstat;

    count = 0;
    errflg = GRR_OK;

    ITR_Chanstat->remove->all();

    for (count = 0; count < nChips; count ++) {
      chanstat = ITR_Chanstat->add->key ((unsigned short) cstat[count].chipid);
      
      if (chanstat == NULL) {
	printf (">>> FAILED to add Chanstat[%d]: 0x%04x\n", count, cstat[count].chipid);
	errflg = GRR_FISHY;
	break;
      }

      for (i = 0; i < NCHIPCHAN; i++) 
	chanstat->status[i] = cstat[count].status[i];
    }

    return (errflg);
}


int ITR_leda_Ped_tableFromStruct (itrleda_Ped_t *cstat, int nChips) {
    /* static char *routine = "ITR_leda_Chip_tableFromStruct"; */
    int count, errflg, i;
    LedaTableRow(ITR_Ped) *ped;

    count = 0;
    errflg = GRR_OK;

    ITR_Ped->remove->all();

    for (count = 0; count < nChips; count ++) {
      ped = ITR_Ped->add->key ((unsigned short) cstat[count].chipid);
      
      if (ped == NULL) {
	printf (">>> FAILED to add Ped[%d]: 0x%04x\n", count, cstat[count].chipid);
	errflg = GRR_FISHY;
	break;
      }

      for (i = 0; i < NCHIPCHAN; i++) 
	ped->pedestal[i] = cstat[count].pedestal[i];


    }

    return (errflg);
}


int ITR_leda_Noise_tableFromStruct (itrleda_Noise_t *cstat, int nChips) {
    /* static char *routine = "ITR_leda_Noise_tableFromStruct"; */
    int count, errflg, i;
    LedaTableRow(ITR_Noise) *noise;

    count = 0;
    errflg = GRR_OK;

    ITR_Noise->remove->all();

    for (count = 0; count < nChips; count ++) {
      noise = ITR_Noise->add->key ((unsigned short) cstat[count].chipid);
      
      if (noise == NULL) {
	printf (">>> FAILED to add Noise[%d]: 0x%04x\n", count, cstat[count].chipid);
	errflg = GRR_FISHY;
	break;
      }

      for (i = 0; i < NCHIPCHAN; i++) {
	noise->subvar[i] = cstat[count].subvar[i];
	noise->rawvar[i] = cstat[count].rawvar[i];
      }

    }
    return (errflg);
}


int ITR_leda_Fed_tableFromStruct (itrleda_Fed_t *fed, int nFeds) {
    /* static char *routine = "ITR_leda_Fed_tableFromStruct"; */
    int count, errflg;
    LedaTableRow(ITR_Fed) *ledaFed;

    count = 0;
    errflg = GRR_OK;

    ITR_Fed->remove->all();

    for (count = 0; count < nFeds; count ++) {
      ledaFed = ITR_Fed->add->key ((unsigned short) fed[count].id);
      
      if (ledaFed == NULL) {
	printf (">>> FAILED to add Fed[%d]: 0x%04x\n", count, ledaFed[count].id);
	errflg = GRR_FISHY;
	break;
      }

      ledaFed->vmeid= fed[count].vmeid;
      ledaFed->board= fed[count].board;
      ledaFed->sharc= fed[count].sharc;
      ledaFed->link = fed[count].link ;

    }
    return (errflg);
}




int ITR_leda_Align_tableFromStruct (itrleda_Align_t *align, int nAlign) {
    /* static char *routine = "ITR_leda_Align_tableFromStruct"; */
    int count, errflg, i;
    LedaTableRow(ITR_Align) *ledaAlign;

    count = 0;
    errflg = GRR_OK;

    ITR_Align->remove->all();

    for (count = 0; count < nAlign; count ++) {
      ledaAlign = ITR_Align->add->key ((unsigned short) align[count].id);
      
      if (ledaAlign == NULL) {
	printf (">>> FAILED to add Align %d 0x%04x\n", count, align[count].id);
	errflg = GRR_FISHY;
	break;
      }
      
      ledaAlign->id     = align[count].id;
      ledaAlign->redres = align[count].redres;
      
      for (i = 0; i < 3; i++) {
	ledaAlign->ang[i]    = align[count].ang[i]   ;
	ledaAlign->eang[i]   = align[count].eang[i]  ;
	ledaAlign->delt[i]   = align[count].delt[i]  ;
	ledaAlign->edelt[i]  = align[count].edelt[i] ;
	ledaAlign->sloff[i]  = align[count].sloff[i] ;
	ledaAlign->esloff[i] = align[count].esloff[i];
      }
      
    }
    return (errflg);
}


int ITR_leda_nede_tableFromStruct (itrleda_nede_t *nede, int nNede) {
    /* static char *routine = "ITR_leda_nede_tableFromStruct"; */
    int count, errflg, i;
    LedaTableRow(ITR_nede) *ledaNede;

    count = 0;
    errflg = GRR_OK;

    ITR_nede->remove->all();

    for (count = 0; count < nNede; count ++) {
      ledaNede = ITR_nede->add->key (nede[count].superl_db, 
				     nede[count].lay_db,
				     nede[count].sec_db[0]);
      
      if (ledaNede == NULL) {
	printf (">>> FAILED to add Nede[%d] %d\n", count, nede[count].inede);
	errflg = GRR_FISHY;
	break;
      }

      ledaNede->inede   = nede[count].inede  ;


      for (i = 0; i < 2; i++) {
	ledaNede->xfw_db[i]  = nede[count].xfw_db[i]  ;
	ledaNede->xmin_db[i] = nede[count].xmin_db[i] ;
	ledaNede->xmax_db[i] = nede[count].xmax_db[i] ;
	ledaNede->ymin_db[i] = nede[count].ymin_db[i] ;
	ledaNede->ymax_db[i] = nede[count].ymax_db[i] ;
	ledaNede->sxmn_db[i] = nede[count].sxmn_db[i] ;
	ledaNede->sxmx_db[i] = nede[count].sxmx_db[i] ;
	ledaNede->symn_db[i] = nede[count].symn_db[i] ;
	ledaNede->symx_db[i] = nede[count].symx_db[i] ;
	ledaNede->sec_db[i]  = nede[count].sec_db[i]  ;
      }
      for (i = 0; i < 9; i++)
	ledaNede->rot_db[i] = nede[count].rot_db[i];
	
      ledaNede->zfw_db  = nede[count].zfw_db ;
      ledaNede->zmin_db = nede[count].zmin_db;
      ledaNede->zmax_db = nede[count].zmax_db;
      ledaNede->acc_db  = nede[count].acc_db ;
      ledaNede->eff_db  = nede[count].eff_db ;
      ledaNede->dod_db  = nede[count].dod_db ;
      ledaNede->pit_db  = nede[count].pit_db ;
      ledaNede->ms_db   = nede[count].ms_db  ;
      ledaNede->nsl_db  = nede[count].nsl_db ;
      ledaNede->lay_db  = nede[count].lay_db ;
      ledaNede->dedx_db = nede[count].dedx_db;
      ledaNede->noise_db= nede[count].noise_db;
      ledaNede->ares_db = nede[count].ares_db;
      ledaNede->superl_db = nede[count].superl_db ;

      strncpy (ledaNede->hwc_db, nede[count].hwc_db, 12);
      
    }
    return (errflg);
}



int ITR_leda_nesl_tableFromStruct (itrleda_nesl_t *nesl, int nNesl) {
    /* static char *routine = "ITR_leda_nesl_tableFromStruct"; */
    int count, errflg;
    LedaTableRow(ITR_nesl) *ledaNesl;

    count = 0;
    errflg = GRR_OK;

    ITR_nesl->remove->all();

    for (count = 0; count < nNesl; count ++) {
      ledaNesl = ITR_nesl->add->key (nesl[count].nsl);
      
      if (ledaNesl == NULL) {
	printf (">>> FAILED to add Nesl[%d] %d\n", count, nesl[count].nsl);
	errflg = GRR_FISHY;
	break;
      }

      ledaNesl->inesl = nesl[count].inesl  ;

      ledaNesl->nsl = nesl[count].nsl;
      ledaNesl->mfl = nesl[count].mfl;
      ledaNesl->lay = nesl[count].lay;
      ledaNesl->x1  = nesl[count].x1 ;
      ledaNesl->x2  = nesl[count].x2 ;
      ledaNesl->y1  = nesl[count].y1 ;
      ledaNesl->y2  = nesl[count].y2 ;
      ledaNesl->z1  = nesl[count].z1 ;
      ledaNesl->z2  = nesl[count].z2 ;
      ledaNesl->dx  = nesl[count].dx ;
      ledaNesl->dy  = nesl[count].dy ;
      ledaNesl->npl = nesl[count].npl;
      ledaNesl->fpl = nesl[count].fpl;
      ledaNesl->ltyp= nesl[count].ltyp;
      ledaNesl->pnede= nesl[count].pnede;

      strncpy (ledaNesl->name, nesl[count].name, 4);
    }
    return (errflg);
}

int ITR_leda_DVOffset_tableFromStruct (itrleda_DVOffset_t *off, int nOffsets) {
    /* static char *routine = "ITR_leda_nesl_tableFromStruct"; */
    int count, errflg;
    LedaTableRow(ITR_DVOffset) *ledaDVOffset;

    count = 0;
    errflg = GRR_OK;

    ITR_DVOffset->remove->all();

    for (count = 0; count < nOffsets; count ++) {
      ledaDVOffset = ITR_DVOffset->add->key (off[count].chamber);
      
      if (ledaDVOffset == NULL) {
	printf (">>> FAILED to add DVOffset[%d] %d\n", count, off[count].chamber);
	errflg = GRR_FISHY;
	break;
      }

      ledaDVOffset->offset = off[count].offset  ;
    }
    return (errflg);
}


/*----------------------------------------------------------------------
 *  build LEDA relations from entries in structure
 *----------------------------------------------------------------------*/

int ITR_leda_RelationsFromStruct (itrleda_fmtTable_t *itrleda_tables) {
  int ichamber, ichip, inede, ifed;
  short int chip_id, chamber_id;
  char *routine = __PRETTY_FUNCTION__;


  aframe_mess (routine, GRR_SINFO, "Relating Chambers with Chips");
  for (ichamber = 0; ichamber < itrleda_tables->nChambers; ichamber ++) {
    chamber_id = itrleda_tables -> chambers[ichamber].id;

    if (itrleda_tables->chambers[ichamber].chip[0] == kNotConnected) 
      continue;

    /* Chamber <-> Chip relation (1 : N) */
    for (ichip = 0; ichip < 6; ichip ++) {
      chip_id = itrleda_tables -> chambers[ichamber].chip[ichip];

      RelateRows (ITR_Chamber, chip, key((unsigned short) chamber_id),
		  ITR_Chip, chamber,  key((unsigned short) chip_id));
    }
  }


  /* Chip <-> {Chanstat, Ped, Noise} relations (1 : 1) */
  aframe_mess (routine, GRR_SINFO, "Relating Chips with ChipInfo");
  for (ichip = 0; ichip < itrleda_tables->nChips; ichip ++) {
    chip_id = itrleda_tables->chips[ichip].id;
    RelateRows (ITR_Chip, chanstat, key((unsigned short) chip_id),
		ITR_Chanstat, chip, key((unsigned short) chip_id));

    RelateRows (ITR_Chip, noise, key((unsigned short) chip_id),
		ITR_Noise, chip, key((unsigned short) chip_id));

    RelateRows (ITR_Chip, pedestals, key((unsigned short) chip_id),
		ITR_Ped, chip, key((unsigned short) chip_id));
  }


  /* nede    <-> nesl relations (N : 1) 
   * Chamber <-> nede relation  (1 : 1)
   * Align   <-> nede relation  (1 : 1)
   */
  aframe_mess (routine, GRR_SINFO, "Relating Geometry, Chambers, and Alignment");
  for (inede = 0; inede < itrleda_tables->nNede; inede ++ ) {
    RelateRows (ITR_nede, tonesl, key(itrleda_tables->nede[inede].superl_db,
				      itrleda_tables->nede[inede].lay_db,
				      itrleda_tables->nede[inede].sec_db[0]),
		ITR_nesl, tonede, key (itrleda_tables->nede[inede].superl_db));

    if (itrleda_tables->nede[inede].toalign != kNotConnected) {
#ifdef DEBUG      
      aframe_mess (routine, GRR_SINFO, "Relating Align %x with NEDE (%d,%d,%d)", itrleda_tables->nede[inede].toalign,
		   itrleda_tables->nede[inede].superl_db,
		   itrleda_tables->nede[inede].lay_db,
		   itrleda_tables->nede[inede].sec_db[0]);
#endif
      RelateRows (ITR_Align, tonede, key(itrleda_tables->nede[inede].toalign),
		  ITR_nede, toalign, key(itrleda_tables->nede[inede].superl_db,
					 itrleda_tables->nede[inede].lay_db,
					 itrleda_tables->nede[inede].sec_db[0]));
    }
	  
    if (itrleda_tables->nede[inede].tochamber != kNotConnected) {
#ifdef DEBUG
      aframe_mess (routine, GRR_SINFO, "Relating Chamber %x with NEDE (%d,%d,%d)", itrleda_tables->nede[inede].tochamber,
		   itrleda_tables->nede[inede].superl_db,
		   itrleda_tables->nede[inede].lay_db,
		   itrleda_tables->nede[inede].sec_db[0]);
#endif
      RelateRows (ITR_Chamber, tonede, key((unsigned short)itrleda_tables->nede[inede].tochamber),
		  ITR_nede, tochamber, key(itrleda_tables->nede[inede].superl_db,
					   itrleda_tables->nede[inede].lay_db,
					   itrleda_tables->nede[inede].sec_db[0]));
    }
    
  }

  
  /* Fed <-> Chip relation (1 : 2) */
  aframe_mess (routine, GRR_SINFO, "Relating Chips with FEDs");
  for (ifed = 0; ifed < itrleda_tables->nFeds; ifed ++) {
    for (ichip = 0; ichip < 2; ichip ++) {

      if (itrleda_tables->fed[ifed].chip[ichip] == kNotConnected)
	continue; 

      RelateRows (ITR_Fed, chip, key ((unsigned short)itrleda_tables->fed[ifed].id),
		  ITR_Chip, fed, key ((unsigned short)itrleda_tables->fed[ifed].chip[ichip]));
    }
  }
  

  return (GRR_OK);
}
