/***********************************************************************
 * 
 *   itrledalib_dbase.c
 *
 *   Description:
 *      More utility functions for itrleda:  Read single tables
 *      needed for the CIF formatter
 *      ITR_leda_fetch_alignment() is used also elsewhere
 *      
 *   Author:   Wolfgang Gradl <gradl@mail.desy.de> 
 *
 *   Changes:
 *      18/oct/2001 (wg)
 *          Add some debugging helps and safety checks
 *
 *   $Id: itrledalib_dbase.c,v 1.1.1.5 2002/02/05 16:31:14 gradl Exp $
 *   
 ***********************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "grr.h"

#include "aframe.h"
#include "aframelib.h"

#include "itrleda.h"
#include "itrleda_dd.h"
#include "itrleda_structs.h"


/* ======================================================================
 * 
 * functions to fetch single tables from a specified data base 
 *
 * ======================================================================*/

int ITR_leda_fetch_alignment (char *database, int align_major, int align_minor) {
  int rc, retries, num_retries, timeout_retries;
  char *routine = "ITR_leda_fetch_alignment";
  static int alignment_fetched = 0;

  MizziDatabase( itrdb_align );

  /* for convenience only */
  num_retries     = itrleda_iol.num_retries.value;
  timeout_retries = itrleda_iol.timeout_retries.value;

  if (alignment_fetched) {
    ITR_Align->remove->all ();
  }
  alignment_fetched ++;

  /* Open Alignment Database  */
  if ( strcmp( database, "none" ) ) { 
    /* Open DB  */ 
    itrdb_align.open( database, 0 );

    /* book tables */
    rc = ITR_Align ->book( itrdb_align );
    if (rc == 0) {
      aframe_mess (routine, GRR_SERROR, "Error booking table ITR_Align!");
      return (GRR_FISHY);
    }

    /* time stamp  */
    aframe_mess(routine,GRR_SINFO, "Reading ITR Alignment DB %s for [%d,%d] ", 
		database, align_major, align_minor); 
    itrdb_align.stamp( align_major, align_minor ); 
    
    /* fetch data */

    retries = 0;
    do {
      if (retries++) { 
	aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_Align.  Retrying... ");
	usleep (timeout_retries);
      }
      rc = ITR_Align->fetch_data( DBeq ); 
    } while (!rc && retries < num_retries);

    if (rc == 0) {
      aframe_mess (routine, GRR_SERROR, "Error reading data for table ITR_Align!");
      return (GRR_FISHY);
    }
         
    /* set keys */
    ITR_Align ->set_keys(); 

    rc = ITR_Align   ->fetch_relations( DBeq );
    
    ITR_Align   ->unbook(); 
    itrdb_align.close(); 
  }   

  return (GRR_OK);
}





int itrleda_fetchtable (char *tablename, int major, int minor) {
  char *routine = "itrleda_fetchtable";
  int rc;
  static int fetchtable_done = 0;
  static int oldmajor = -1, oldminor = -1;
  static int oldmajor_align = -1, oldminor_align = -1;
  static int oldmajor_mon   = -1, oldminor_mon   = -1;
  static int oldmajor_off   = -1, oldminor_off   = -1;   /* for the DV offsets table */

  extern Itrleda_Iol_t itrleda_iol;

  /* declare databases */
  MizziDatabase( itrdb_setup );
  MizziDatabase( itrdb_mon   );

  /* Check if we haven't called it yet */
  if ( !fetchtable_done ) {

  }

  fetchtable_done = 1;

  /* Open Setup Database */
  if ( strcmp( itrleda_iol.setup_database.value, "none" ) ) {
    /* open DB */
    itrdb_setup.open( itrleda_iol.setup_database.value, 0 );

    /* book tables 
     *
     * book() returns 0 on failure and >0 on success 
     */   
    rc = 0;
    
    rc |= ! ITR_Chip    ->book( itrdb_setup );
    rc |= ! ITR_Fed     ->book( itrdb_setup );
    rc |= ! ITR_Chamber ->book( itrdb_setup );
    rc |= ! ITR_nesl    ->book( itrdb_setup );
    rc |= ! ITR_nede    ->book( itrdb_setup );
    rc |= ! ITR_TFU     ->book( itrdb_setup );
    rc |= ! ITR_FedSlb  ->book( itrdb_setup );
    rc |= ! ITR_DVOffset->book( itrdb_setup );

    if (rc) {
      aframe_mess(routine, GRR_SERROR, 
		  "Error booking tables in server %s!", itrleda_iol.setup_database.value);
      return (GRR_FISHY);
    }
  }

  /* Open Monitoring Database */
  if ( strcmp( itrleda_iol.mon_database.value, "none" ) ) {
  
    /* Open DB */
    itrdb_mon.open( itrleda_iol.mon_database.value, 0 );
    
    rc = 0;
    rc |= ! ITR_Chanstat->book( itrdb_mon );
    rc |= ! ITR_Ped     ->book( itrdb_mon );
    rc |= ! ITR_Noise   ->book( itrdb_mon );

    if (rc) {
      aframe_mess(routine, GRR_SERROR, 
		  "Error booking tables in server %s!", itrleda_iol.mon_database.value);
      return (GRR_FISHY);
    }
  }


  if((!strcmp (tablename, "ITR_Chip") ||
      !strcmp (tablename, "ITR_Fed")  || 
      !strcmp (tablename, "ITR_Chamber") ||
      !strcmp (tablename, "ITR_nesl") ||
      !strcmp (tablename, "ITR_nede") ) ) {

    if (major != oldmajor || minor != oldminor) {
      rc  = itrleda_fetch_Chip (&itrdb_setup, major, minor);
      rc |= itrleda_fetch_Fed  (&itrdb_setup, major, minor);
      rc |= itrleda_fetch_Chamber (&itrdb_setup, major, minor);
      rc |= itrleda_fetch_nesl (&itrdb_setup, major, minor);
      rc |= itrleda_fetch_nede (&itrdb_setup, major, minor);
      
      if (rc == GRR_OK) {
	ITR_Chamber->fetch_relations(DBeq);
	ITR_Chip->fetch_relations (DBeq);
	ITR_Fed->fetch_relations (DBeq);
	ITR_nesl->fetch_relations (DBeq);
	ITR_nede->fetch_relations (DBeq);
	
	oldmajor = major;
	oldminor = minor;
      }
    }
  }

  else if (!strcmp (tablename, "ITR_DVOffset")) {
    rc = itrleda_fetch_DVOffset (&itrdb_setup, major, minor);
    oldmajor_off = major;
    oldminor_off = minor;
  }

  else if (!strcmp (tablename, "ITR_Ped")) {
    rc = itrleda_fetch_Ped (&itrdb_mon, major, minor);
    ITR_Ped->fetch_relations (DBeq);
    oldmajor_mon = major;
    oldminor_mon = minor;
  }
  else if (!strcmp (tablename, "ITR_Noise")) {
    rc = itrleda_fetch_Noise (&itrdb_mon, major, minor);
    ITR_Noise->fetch_relations (DBeq);
    oldmajor_mon = major;
    oldminor_mon = minor;
  }
  else if (!strcmp (tablename, "ITR_Chanstat")) {
    rc = itrleda_fetch_Chanstat (&itrdb_mon, major, minor);
    ITR_Chanstat->fetch_relations (DBeq);
    oldmajor_mon = major;
    oldminor_mon = minor;
 }      
  
  else if (!strcmp (tablename, "ITR_Align")) {
    rc = ITR_leda_fetch_alignment ("/ITR_Align", major, minor);
    oldmajor_align = major;
    oldminor_align = minor;
  }

  else {
    aframe_mess (routine, GRR_SERROR, "Unknown table %s requested!", tablename);
  }    


  ITR_Chip    ->unbook();
  ITR_Fed     ->unbook();
  ITR_Chamber ->unbook();
  ITR_nede    ->unbook();
  ITR_nesl    ->unbook();
  ITR_TFU     ->unbook();
  ITR_FedSlb  ->unbook();
  ITR_DVOffset->unbook();
  itrdb_setup.close();

  ITR_Ped     ->unbook();
  ITR_Noise   ->unbook();    
  ITR_Chanstat->unbook();
  itrdb_mon.close();

  return rc;
}




int itrleda_fetch_Chip (struct __MizziDataBase *db, int major, int minor) {
  int rc, retries, num_retries, timeout_retries;
  int ITR_Chip_major, ITR_Chip_minor;
  char *routine = __PRETTY_FUNCTION__;

  /* for convenience only */
  num_retries     = itrleda_iol.num_retries.value;
  timeout_retries = itrleda_iol.timeout_retries.value;

  ITR_Chip->remove->all();

  db->stamp (major, minor);

  retries = 0;
  do {
    if (retries++) { 
      aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_Chip.  Retrying... ");
      usleep (timeout_retries);
    }
    rc = ITR_Chip->fetch_data( DBeq ); 
  } while (!rc && retries < num_retries);

  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_Chip!");
    return (GRR_FISHY);
  }

  ITR_Chip->set_keys();
  rc = ITR_Chip->match( DBeq );
  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error matching table ITR_Chip!");
    return (GRR_FISHY);
  }
    
  ITR_Chip_major = TableMajor( db->get_id() );
  ITR_Chip_minor = TableMinor( db->get_id() );
    
  /*   if ( itrleda_iol.debug_mode.value ) */
  aframe_mess(routine,GRR_SINFO, "ITR_Chip read for %s, [%d,%d]",
		print_date(ITR_Chip_major), ITR_Chip_major, ITR_Chip_minor );

  return GRR_OK;
}



int itrleda_fetch_Fed (struct __MizziDataBase *db, int major, int minor) {
  int rc, retries, num_retries, timeout_retries;
  int ITR_Fed_major, ITR_Fed_minor;
  char *routine = __PRETTY_FUNCTION__;

  /* for convenience only */
  num_retries     = itrleda_iol.num_retries.value;
  timeout_retries = itrleda_iol.timeout_retries.value;

  ITR_Fed->remove->all();

  db->stamp (major, minor);

  retries = 0;
  do {
    if (retries++) { 
      aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_Fed.  Retrying... ");
      usleep (timeout_retries);
    }
    rc = ITR_Fed->fetch_data( DBeq ); 
  } while (!rc && retries < num_retries);

  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_Fed!");
    return (GRR_FISHY);
  }
  ITR_Fed->set_keys();
  rc = ITR_Fed->match( DBeq );
  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error matching table ITR_Fed!");
    return (GRR_FISHY);
  }
  ITR_Fed_major = TableMajor( db->get_id() );
  ITR_Fed_minor = TableMinor( db->get_id() );

  /*     if ( itrleda_iol.debug_mode.value ) */
  aframe_mess(routine,GRR_SINFO, "ITR_Fed read for %s, [%d,%d]",
	      print_date(ITR_Fed_major), ITR_Fed_major, ITR_Fed_minor );
  return GRR_OK;
  
}


int itrleda_fetch_Chamber(struct __MizziDataBase *db, int major, int minor) {
  int rc, retries, num_retries, timeout_retries;
  int ITR_Chamber_major, ITR_Chamber_minor;
  char *routine = __PRETTY_FUNCTION__;
 
  /* for convenience only */
  num_retries     = itrleda_iol.num_retries.value;
  timeout_retries = itrleda_iol.timeout_retries.value;

  ITR_Chamber->remove->all(); 

  db->stamp (major, minor);

  retries = 0;
  do {
    if (retries++) { 
      aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_Chamber.  Retrying... ");
      usleep (timeout_retries);
    }
    rc = ITR_Chamber->fetch_data( DBeq ); 
  } while (!rc && retries < num_retries);

  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_Chamber!");
    return (GRR_FISHY);
  }

  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_Chamber!");
    return (GRR_FISHY);
  } 
  ITR_Chamber->set_keys();
  rc = ITR_Chamber->match( DBeq );
  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error matching table ITR_Chamber!");
    return (GRR_FISHY);
  }
  ITR_Chamber_major = TableMajor( db->get_id() );
  ITR_Chamber_minor = TableMinor( db->get_id() );

  aframe_mess(routine,GRR_SINFO, "ITR_Chamber read for %s, [%d,%d] ",
	      print_date(ITR_Chamber_major), ITR_Chamber_major, ITR_Chamber_minor );

  return GRR_OK;
}


int itrleda_fetch_nesl(struct __MizziDataBase *db, int major, int minor) {
  int rc, retries, num_retries, timeout_retries;
  char *routine = __PRETTY_FUNCTION__;

  /* for convenience only */
  num_retries     = itrleda_iol.num_retries.value;
  timeout_retries = itrleda_iol.timeout_retries.value;

  ITR_nesl->remove->all(); 

  db->stamp (major, minor);

  retries = 0;
  do {
    if (retries++) { 
      aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_nesl.  Retrying... ");
      usleep (timeout_retries);
    }
    rc = ITR_nesl->fetch_data( DBeq ); 
  } while (!rc && retries < num_retries);


  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_nesl!");
    return (GRR_FISHY);
  }
  ITR_nesl -> set_keys   ();
  rc = ITR_nesl -> match      (DBeq);
  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error matching table ITR_nesl!");
    return (GRR_FISHY);
  }
  return GRR_OK;
}



int itrleda_fetch_nede(struct __MizziDataBase *db, int major, int minor) {
  int rc, retries, num_retries, timeout_retries;
  char *routine = __PRETTY_FUNCTION__;

  /* for convenience only */
  num_retries     = itrleda_iol.num_retries.value;
  timeout_retries = itrleda_iol.timeout_retries.value;

  ITR_nede->remove->all(); 

  db->stamp (major, minor);

  retries = 0;
  do {
    if (retries++) { 
      aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_nede.  Retrying... ");
      usleep (timeout_retries);
    }
    rc = ITR_nede->fetch_data( DBeq ); 
  } while (!rc && retries < num_retries);


  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_nede!");
    return (GRR_FISHY);
  }
  ITR_nede -> set_keys   ();
  rc = ITR_nede -> match      (DBeq);
  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error matching table ITR_nede!");
    return (GRR_FISHY);
  }
  return GRR_OK;
}


int itrleda_fetch_DVOffset (struct __MizziDataBase *db, int major, int minor) {
  int rc, retries, num_retries, timeout_retries;
  int ITR_DVOffset_major, ITR_DVOffset_minor;
  char *routine = __PRETTY_FUNCTION__;

  /* for convenience only */
  num_retries     = itrleda_iol.num_retries.value;
  timeout_retries = itrleda_iol.timeout_retries.value;

  ITR_DVOffset->remove->all();

  db->stamp (major, minor);

  retries = 0;
  do {
    if (retries++) { 
      aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_DVOffset.  Retrying... ");
      usleep (timeout_retries);
    }
    rc = ITR_DVOffset->fetch_data( DBeq ); 
  } while (!rc && retries < num_retries);

  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_DVOffset!");
    return (GRR_FISHY);
  }

  ITR_DVOffset->set_keys();
  rc = ITR_DVOffset->match( DBeq );
  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error matching table ITR_DVOffset!");
    return (GRR_FISHY);
  }
    
  ITR_DVOffset_major = TableMajor( db->get_id() );
  ITR_DVOffset_minor = TableMinor( db->get_id() );
    
  /*   if ( itrleda_iol.debug_mode.value ) */
  aframe_mess(routine,GRR_SINFO, "ITR_DVOffset read for %s, [%d,%d]",
		print_date(ITR_DVOffset_major), ITR_DVOffset_major, ITR_DVOffset_minor );

  return GRR_OK;
}



int itrleda_fetch_Chanstat(struct __MizziDataBase *db, int major, int minor) {
  int rc, retries, num_retries, timeout_retries;
  char *routine = __PRETTY_FUNCTION__;

  /* for convenience only */
  num_retries     = itrleda_iol.num_retries.value;
  timeout_retries = itrleda_iol.timeout_retries.value;

  ITR_Chanstat->remove->all(); 

  db->stamp (major, minor);

  retries = 0;
  do {
    if (retries++) { 
      aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_Chanstat.  Retrying... ");
      usleep (timeout_retries);
    }
    rc = ITR_Chanstat->fetch_data( DBeq ); 
  } while (!rc && retries < num_retries);

  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_Chanstat!");
    return (GRR_FISHY);
  }
  ITR_Chanstat -> set_keys   ();
  rc = ITR_Chanstat -> match      (DBeq);
  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error matching table ITR_Chanstat!");
    return (GRR_FISHY);
  }
  return GRR_OK;
}


 
int itrleda_fetch_Noise(struct __MizziDataBase *db, int major, int minor) {
  int rc, retries, num_retries, timeout_retries;
  int ITR_Noise_major, ITR_Noise_minor;
  char *routine = __PRETTY_FUNCTION__;

  /* for convenience only */
  num_retries     = itrleda_iol.num_retries.value;
  timeout_retries = itrleda_iol.timeout_retries.value;

  ITR_Noise->remove->all(); 

  db->stamp (major, minor);

  retries = 0;
  do {
    if (retries++) { 
      aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_Noise.  Retrying... ");
      usleep (timeout_retries);
    }
    rc = ITR_Noise->fetch_data( DBeq ); 
  } while (!rc && retries < num_retries);

  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_Noise!");
    return (GRR_FISHY);
  }
  ITR_Noise -> set_keys   ();
  rc = ITR_Noise -> match      (DBeq);
  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error matching table ITR_Noise!");
    return (GRR_FISHY);
  }
  ITR_Noise_major = TableMajor( db->get_id() );
  ITR_Noise_minor = TableMinor( db->get_id() );

  aframe_mess(routine,GRR_SINFO, "ITR_Noise read for %s, [%d,%d]",
	      print_date(ITR_Noise_major), ITR_Noise_major, ITR_Noise_minor );
  return GRR_OK;
}


int itrleda_fetch_Ped(struct __MizziDataBase *db, int major, int minor) {
  int rc, retries, num_retries, timeout_retries;
  char *routine = __PRETTY_FUNCTION__;

  /* for convenience only */
  num_retries     = itrleda_iol.num_retries.value;
  timeout_retries = itrleda_iol.timeout_retries.value;

  ITR_Ped->remove->all(); 

  db->stamp (major, minor);

  retries = 0;
  do {
    if (retries++) { 
      aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_Ped.  Retrying... ");
      usleep (timeout_retries);
    }
    rc = ITR_Ped->fetch_data( DBeq ); 
  } while (!rc && retries < num_retries);

  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_Ped!");
    return (GRR_FISHY);
  }
  ITR_Ped -> set_keys   ();
  rc = ITR_Ped -> match      (DBeq);
  if (!rc) {
    aframe_mess(routine, GRR_SERROR, "Error matching table ITR_Ped!");
    return (GRR_FISHY);
  }
  return GRR_OK;
}
